var vows = require("vows"),
	Trait = require("traits").Trait;

// utility create function
var create = function(trait) {
	return Object.create(Object.prototype, trait);
};

// An object that can be started
var StartAble = Trait({
	// initialize the object
	"start": function _start() {
		this._val = {};
		return this;
	}
});

// An object that can be set
var SetAble = Trait({
	// sets the value of the object
	"set": function _set(value) {
		this._val = value;
		return this;
	}
});

// An object that can end.
var EndAble = Trait({
	// end the object. This returns the Object that vow expects to be 
	// an object. This will call end recursively if it exists.
	"end": function _end() {
		var value = {}, val = this._val;
		Object.keys(val).forEach(function(key) {
			if (val[key].end) {
				value[key] = val[key].end();
			} else {
				value[key] = val[key];
			}
		});
		return value;
	}
});

// An object which can navigate up it's chain.
var ParentAble = Trait({
	// Returns the parent context
	"parent": function _parent() {
		return this._parent;
	},
	// returns the batch this context belongs to
	"batch": function _batch() {
		return this._batch;
	},
	// returns the suite this context belongs to
	"suite": function _suite() {
		return this._batch._suite;
	}
});

// An object which can create contexts
var ContextAble = Trait({
	// Create a Context and add it to this object
	// with the name `text`. Returns the Context.
	"context": function _context(text) {
		var c = createContext();
		c._parent = this;
		c._batch = this._batch;
		this._val[text] = c;
		return c;
	}
});

// Topic object. This is an abstraction around Vows topics.
// Created for ease of extension development
var Topic = create(Trait.compose(SetAble, ParentAble, EndAble));

// Vow object. This is an abstraction around Vows vows.
// Created for ease of extension development
var Vow = create(Trait.compose(SetAble, ParentAble, EndAble));

// Context object. This is an abstraction around Vows contexts
var Context = create(Trait.compose(Trait({
	// Sets the topic of the context to be `topic`
	// Or returns a Topic if no `topic` was passed.
	"topic": function _topic(topic) {
		if (topic) {
			this._val.topic = topic;
			return this;
		} else {
			var t = Object.create(Topic);
			t._parent = this;
			t._batch = this._batch;
			this._val.topic = t;
			return t;	
		}
	},
	// Adds a vow to the contex
	// with the name `text` and value `vow`
	// or returns a Vow if no `vow` was passed
	"vow": function _vow(text, vow) {
		if (vow) {
			this._val[text] = vow;	
			return this;
		} else {
			var v = Object.create(Vow);
			v._parent = this;
			v._batch = this._batch;
			this._val[text] = v;
			return v;
		}
	}
}), StartAble, EndAble, ParentAble, ContextAble));

// A representation of a vows batch.
// A vows batch just contains a bunch of contexts
var Batch = create(Trait.compose(StartAble, ParentAble, EndAble, ContextAble));

// A representation of a vows Suite. Every application
// should onhly have one suite.
var Suite = create(Trait({
	// initialize the object
	"start": function _start(name) {
		this._val = [];
		this._suite = vows.describe(name);
		return this;
	},
	// Creates a batch and add's it to the suite.
	// Returns the batch
	"batch": function _batch() {
		var b = createBatch();
		b._suite = b._parent = this;
		b._batch = b;
		this._val.push(b);
		return b;
	},
	// end the suite and call export on the vows suite
	"export": function _export(m, options) {
		var s = this.end();
		return s.export(m, options);
	},
	// A macro for JSLint
	"exportTo": function _exportTo(m, options) {
		return this.export(m, options);
	},
	// end the suite and call run on the vows suite
	"run": function _run(options, callback) {
		var s = this.end();
		return s.run(options, callback);	
	},
	// end the suite returning a vows Suite.
	"end": function _end() {
		this._val.forEach((function _addBatches(batch) {
			this._suite.addBatch(batch.end());
		}).bind(this));
		return this._suite;
	}
}));

// creates a batch by cloning `Batch` and 
// starting it.
var createBatch = function() {
	return create(Trait(Batch)).start();
}

// Create a context by cloning `Context` and 
// starting it.
var createContext = function() {
	return create(Trait(Context)).start();
}

// Create a Suite by cloning `Suite` and starting
// it with `name`
var createSuite = function(name) {
	return create(Trait(Suite)).start(name);
}

// expose objects
exports.Vow = Vow;
exports.Topic = Topic
exports.Context = Context;
exports.Batch = Batch;
exports.Suite = Suite;
exports.context = exports.createContext = createContext;
exports.batch = exports.createBatch = createBatch;
exports.suite = exports.createSuite = createSuite;